/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
#include "config.h"
#include "jed-feat.h"
/*{{{ Include Files */

#include <stdio.h>
#include <slang.h>

#include "jdmacros.h"

#include <string.h>
#include <setjmp.h>

#include "buffer.h"
#include "cmds.h"
#include "paste.h"
#include "screen.h"
#include "window.h"
#include "text.h"
#include "ledit.h"
#include "keymap.h"
#include "sysdep.h"
#include "misc.h"
#include "display.h"
#include "file.h"
#include "undo.h"
#include "ins.h"
#include "hooks.h"
#include "replace.h"
#include "indent.h"
#if JED_HAS_MENUS
# include "menu.h"
#endif

#ifdef IBMPC_SYSTEM
# include "doskeys.h"
#endif

////
#include "j2s.h"
int TXTC;
int modifier = 0;
int shiftkey_pressed = 0;
int ctrlkey_pressed = 0;
int shiftctrlkey_pressed = 0;
int altkey_pressed = 0;
//// shift status variables

#if (SLANG_VERSION < 20003) && defined(SLKEY_F_SLANG)
# undef SLKEY_F_SLANG
#endif

void (*X_Define_Keys_Hook)(SLKeyMap_List_Type *);
int (*X_Set_Abort_Char_Hook) (unsigned char);
////unsigned char Jed_Abort_Char = 7;  /* ^G */
unsigned char Jed_Abort_Char = 28;

typedef struct
{
   jmp_buf b;
}

jmp_buf_struct;

jmp_buf_struct Jump_Buffer, *Jump_Buffer_Ptr;

FVOID_STAR Last_Key_Function;
FVOID_STAR Jed_This_Key_Function;

int *Repeat_Factor;                    /* repeats a key sequence if non null */

char *Read_This_Character = NULL;      /* alternate keyboard buffer */

char Jed_Key_Buffer[SLANG_MAX_KEYMAP_KEY_SEQ + 1];
char Key_Buffer[SLANG_MAX_KEYMAP_KEY_SEQ + 1];
char *Key_Bufferp = Key_Buffer;
static int Last_Key_Buffer_Len;

int Jed_Max_Hits = 300;

int jed_beep (void)
{
   //tt_beep ();
   flush_output ();
   return 1;
}

static int redraw (void)
{
   ////jed_redraw_screen (0);
   jed_redraw_screen (1);
   return 0;
}

#if HAS_MOUSE
#ifndef IBMPC_SYSTEM
static int xterm_mouse_cmd (void)
{
   int b, id;
   static int last_down = JMOUSE_BUTTON_1;
   JMouse_Type m;

   b = my_getkey ();
   m.x = (unsigned char) my_getkey () - 32;
   m.y = (unsigned char) my_getkey () - 32;

   b -= 32;

   if ((b & 3) == 3)
     {
        m.type = JMOUSE_UP;
        m.button = last_down;
        m.state = last_down;
     }
   else
     {
        m.type = JMOUSE_DOWN;
        if ((b & 3) == 0)
          m.button = JMOUSE_BUTTON_1;
        else if (b & 1)
          m.button = JMOUSE_BUTTON_2;
        else if (b & 2)
          m.button = JMOUSE_BUTTON_3;
        m.state = 0;
     }
   last_down = m.button;

   if (-1 == (id = jed_mouse_add_event (&m)))
     return -1;

   ungetkey (&id);
   jed_mouse_cmd ();
   return 1;
}
#endif
#endif                                 /* HAS_MOUSE */

SLKeymap_Function_Type Jed_Functions[] =
{   //// 1st element in pair is used in key definitions
   /* Put common cases here to speedup lookup */
     {"self_insert_cmd", ins_char_cmd},
     {"next_char_cmd", next_char_cmd},
     {"next_line_cmd", next_line_cmd},
     {"previous_char_cmd", previous_char_cmd},
     {"previous_line_cmd", previous_line_cmd},
     {"delete_char_cmd", delete_char_cmd},
     {"beg_of_line", bol},
     {"eol_cmd", eol_cmd},
     {"beg_of_buffer", bob},
     {"end_of_buffer", eob},
     {"newline_and_indent", newline_and_indent},
     {"newline", newline_cmd},
     {"page_down", pagedown_cmd},
     {"page_up", pageup_cmd},
     {"backward_delete_char", backward_delete_char_cmd},
     {"backward_delete_char_untabify", backward_delete_char_untabify},
     ////{"beep", jed_beep},
     {"begin_macro", begin_keyboard_macro},
     {"center_line", center_line},
     {"copy_region", copy_to_pastebuffer},
#if (defined(__MSDOS__) && !defined(__WIN32__)) || (defined(__QNX__) && defined(__WATCOMC__))
     {"coreleft", show_memory},
#endif
     {"delete_window", delete_window},
     ////{"digit_arg", digit_arg},
/*      {"double_line", double_line}, */
     {"end_macro", end_keyboard_macro},
     ////{"enlarge_window", enlarge_window},
     {"evaluate_cmd", evaluate_cmd},
     {"execute_macro", execute_keyboard_macro},
     {"exchange",exchange_point_mark},
     {"exit_jed", jed_exit_jed_cmd},
     {"exit_mini", exit_minibuffer},
     {"find_file", find_file},
     {"format_paragraph", text_format_paragraph},
     ////{"goto_match", goto_match},
     {"indent_line", indent_line},
     {"insert_file", insert_file_cmd},
#ifdef HAS_MOUSE
     {"mouse_cmd", jed_mouse_cmd},
#endif
     {"kbd_quit", kbd_quit},
     ////{"kill_buffer", kill_buffer},
     ////{"kill_line", kill_line},
     {"kill_region", kill_region},
     {"macro_query", macro_query},
     {"mark_spot", mark_spot},
     {"mini_complete", mini_complete},
     ////{"narrow", narrow_to_lines},
     ////{"narrow_paragraph", narrow_paragraph},
     ////{"narrow_to_region", narrow_to_region},
     {"one_window", one_window},
     {"other_window", other_window},
     {"pop_spot", pop_spot},
     {"quoted_insert", quoted_insert},
     {"redraw", redraw},
      /* {"replace", replace}, */
     {"save_some_buffers", save_some_buffers},
     {"scroll_left", jed_scroll_left_cmd},
     {"scroll_right", jed_scroll_right_cmd},
       /* {"search_backward", search_backward_cmd},
          {"search_forward", search_forward_cmd}, */
     {"set_mark_cmd", set_mark_cmd},
     {"split_window", split_window},
     {"switch_to_buffer", get_buffer},
     {"sys_spawn_cmd", sys_spawn_cmd},
     {"text_smart_quote", text_smart_quote},
/*       {"transpose_lines", transpose_lines}, */
     {"trim_whitespace", jed_trim_whitespace},
     {"undo", undo},
     ////{"widen", widen},
     ////{"widen_region", widen_region},
     ////{"write_buffer", jed_save_buffer_as_cmd},
     {"yank", yank},
#if JED_HAS_MENUS
     {"select_menubar", jed_select_menu_bar},
#endif
     {(char *) NULL, NULL}
};

SLKeyMap_List_Type *Global_Map;

////
SLKeyMap_List_Type *Slim_Map;
SLKeyMap_List_Type *S_Map;
SLKeyMap_List_Type *C_Map;
SLKeyMap_List_Type *S_C_Map;
SLKeyMap_List_Type *A_Map;
////  keymaps for extended keys and slim keymap


////
int kbd_quit(void)
{
    int ignore_quit = 0;

    if (Ignore_User_Abort)
        ignore_quit=1;

    #if JED_HAS_MENUS
    else if (Jed_Menus_Active)
        ignore_quit=1;
    #endif

    if(ignore_quit)
    {
        *Error_Buffer = 0;
        SLKeyBoard_Quit = 0;
        SLang_set_error (0);
        return 1;
    }

    SLKeyBoard_Quit = 1;
    SLang_set_error (SL_USER_BREAK);

    #ifdef UNDO_HAS_REDO
    set_current_undo ();
    #endif

    return 1;
}


void init_keymaps(void)
{
    #include "keys.h"
    //// keys names

   int ch;
   char simple[2];
   simple[1] = 0;

   if (NULL == (Global_Map = SLang_create_keymap ("global", NULL)))
     exit_error ("Malloc error creating keymap.", 0);

   Global_Map->functions = Jed_Functions;

   ////
    //  repeat Global_Map definition for ctrl_kmap and so on
    //  then specify just a couple of keys for each
    //  they will be customized with definekey (...)
    if (NULL == (S_Map = SLang_create_keymap ("shift_kmap", NULL)))
        exit_error ("Malloc error creating keymap.", 0);
    SLkm_define_key ("^M", (FVOID_STAR) newline, S_Map);
    SLkm_define_key ("\r", (FVOID_STAR) newline, S_Map);

    if (NULL == (C_Map = SLang_create_keymap ("ctrl_kmap", NULL)))
        exit_error ("Malloc error creating keymap.", 0);
    SLkm_define_key (Key_Home, (FVOID_STAR) bob, C_Map);
    SLkm_define_key (Key_End, (FVOID_STAR) eob, C_Map);

    if (NULL == (S_C_Map = SLang_create_keymap ("shiftctrl_kmap", NULL)))
        exit_error ("Malloc error creating keymap.", 0);
    SLkm_define_key (Key_Home, (FVOID_STAR) bob, S_C_Map);
    SLkm_define_key (Key_End, (FVOID_STAR) eob, S_C_Map);

    if (NULL == (A_Map = SLang_create_keymap ("alt_kmap", NULL)))
        exit_error ("Malloc error creating keymap.", 0);
    SLkm_define_key ("-", (FVOID_STAR) center_line, A_Map);
    SLkm_define_key ("+", (FVOID_STAR) center_line, A_Map);

    if (NULL == (Slim_Map = SLang_create_keymap ("slim_kmap", NULL)))
        exit_error ("Malloc error creating keymap.", 0);

    Slim_Map->functions = Jed_Functions;
    for (ch = 32; ch < 256; ch++)
    {
        if (ch==127) continue;
        simple[0] = (char) ch;
        SLkm_define_key (simple, (FVOID_STAR) ins_char_cmd, Slim_Map);
        /////////////
    }

    SLkm_define_key ("^M",      (FVOID_STAR) newline_and_indent, Slim_Map);

    SLkm_define_key (Key_Right, (FVOID_STAR) next_char_cmd, Slim_Map);
    SLkm_define_key (Key_Left,  (FVOID_STAR) previous_char_cmd, Slim_Map);

    SLkm_define_key (Key_Up,    (FVOID_STAR) previous_line_cmd, Slim_Map);
    SLkm_define_key (Key_Down,  (FVOID_STAR) next_line_cmd, Slim_Map);

    SLkm_define_key (Key_Home,  (FVOID_STAR) bol, Slim_Map);
    SLkm_define_key (Key_End,   (FVOID_STAR) eol, Slim_Map);

    SLkm_define_key (Key_Del,   (FVOID_STAR) delete_char_cmd, Slim_Map);
    SLkm_define_key (Key_BS,    (FVOID_STAR) backward_delete_char_cmd, Slim_Map);
    SLkm_define_key ("^\\",     (FVOID_STAR) kbd_quit, Slim_Map);
    ////

   /* This breaks under some DEC ALPHA compilers (scary!) */
#ifndef __DECC
   for (ch = ' '; ch < 256; ch++)
     {
        simple[0] = (char) ch;
        SLkm_define_key (simple, (FVOID_STAR) ins_char_cmd, Global_Map);
     }
#else
   ch = ' ';
   while (1)
     {
        simple[0] = (char) ch;
        SLkm_define_key (simple, (FVOID_STAR) ins_char_cmd, Global_Map);
        ch = ch + 1;
        if (ch == 256) break;
     }
#endif

    //// Global_Map greatly reduced and modified

   SLkm_define_key (Key_Left,  (FVOID_STAR) previous_char_cmd, Global_Map);
   SLkm_define_key (Key_Up,    (FVOID_STAR) previous_line_cmd, Global_Map);
   SLkm_define_key (Key_Down,  (FVOID_STAR) next_line_cmd, Global_Map);
   SLkm_define_key (Key_Right, (FVOID_STAR) next_char_cmd, Global_Map);
   SLkm_define_key (Key_PgDn,  (FVOID_STAR) pagedown_cmd, Global_Map);
   SLkm_define_key (Key_PgUp,  (FVOID_STAR) pageup_cmd, Global_Map);

   SLkm_define_key (Key_Home,  (FVOID_STAR) bol, Global_Map);
   SLkm_define_key (Key_End,   (FVOID_STAR) eol_cmd, Global_Map);

   SLkm_define_key (Key_Del,   (FVOID_STAR) delete_char_cmd, Global_Map);
   SLkm_define_key (Key_BS,    (FVOID_STAR) backward_delete_char_cmd, Global_Map);

   SLkm_define_key (Key_F10,   (FVOID_STAR) jed_select_menu_bar, Global_Map);
   SLkm_define_key ("^O",      (FVOID_STAR) find_file, Global_Map);

   SLkm_define_key ("^\\",     (FVOID_STAR) kbd_quit, Global_Map);
   SLkm_define_key ("^KE",     (FVOID_STAR) jed_exit_jed_cmd, Global_Map);
   SLkm_define_key ("^M",      (FVOID_STAR) newline_and_indent, Global_Map);
    ////

   if (X_Define_Keys_Hook != NULL)  (*X_Define_Keys_Hook)(Global_Map);
}

char Last_Kbd_Command_String[32];
char Current_Kbd_Command_String[32];

void set_current_kbd_command (char *s)
{
   ////strncpy (Current_Kbd_Command_String, s, 31);
   ////Current_Kbd_Command_String[31] = 0;

    //  save current command only from normal editing buffers

    char * nada="nada";

    if (slim_test())
    {
        strncpy (Current_Kbd_Command_String, nada, 4);
        Current_Kbd_Command_String[4] = 0;
    }

    else if (NULL != strstr(CBuf->name, "/"))
    {
        strncpy (Current_Kbd_Command_String, nada, 4);
        Current_Kbd_Command_String[4] = 0;
    }

    else if (strlen(s) > 1)
    {
        strncpy (Current_Kbd_Command_String, s, 31);
        Current_Kbd_Command_String[31] = 0;
    }
}

static void update_jed_keybuffer (void)
{
   Last_Key_Buffer_Len = (int) (Key_Bufferp - Key_Buffer);
   if (Last_Key_Buffer_Len == 1)
     {
        *Jed_Key_Buffer = *Key_Buffer;
        *(Jed_Key_Buffer + 1) = 0;
     }
   else
     {
        *Key_Bufferp = 0;
        strcpy (Jed_Key_Buffer, Key_Buffer);
     }

   Key_Bufferp = Key_Buffer;
}

static int do_macro_string (char *str, int repeat)
{
   static int macro_depth;
   char *rtc_save;
   int ret = 0;

   if (macro_depth > 10)
     {
        macro_depth = 0;
        msg_error ("Possible runaway macro aborted.");
        return -1;
     }

   macro_depth++;
   rtc_save = Read_This_Character;

   while (repeat > 0)
     {
        repeat--;

        Read_This_Character = str;

        while ((SLang_get_error () == 0)
               && (SLKeyBoard_Quit == 0)
               && (Read_This_Character != NULL))
          jed_do_key();

        if (Read_This_Character != NULL)
          {
             ret = -1;
             break;
          }
     }

   Read_This_Character = rtc_save;
   macro_depth--;

   return ret;
}

static int key_interpret (SLang_Key_Type *key)
{
   char *str;
   int ret;

   strcpy (Last_Kbd_Command_String, Current_Kbd_Command_String);
   Jed_This_Key_Function = key->f.f;

   switch (key->type)
     {
      case SLKEY_F_INTRINSIC:
        set_current_kbd_command (Jed_Key_Buffer);
        ret = (key->f.f) ();
        break;

      case SLKEY_F_INTERPRET:
        str = key->f.s;
        set_current_kbd_command (str);

        if (*str == ' ')
          {
             /* string to insert */
             (void) jed_insert_string(str + 1);
          }
        else if (*str == '@')
          {
             int repeat;

             if (Repeat_Factor != NULL)
               {
                  repeat = *Repeat_Factor;
                  Repeat_Factor = NULL;
               }
             else repeat = 1;

             (void) do_macro_string (str + 1, repeat);
          }
        else if ((*str == '.')
                 || !SLang_execute_function(str))
          SLang_load_string(str);

        ret = 1;
        break;

#ifdef SLKEY_F_SLANG
      case SLKEY_F_SLANG:
        ret = SLexecute_function (key->f.slang_fun);
        break;
#endif

      default:
        SLang_verror (SL_NOT_IMPLEMENTED, "Keymap key-type %d is not supported", (int)key->type);
        return -1;
     }

   Last_Key_Function = Jed_This_Key_Function;
   return(ret);
}

////
int get_modifier ()
///////////////////////////////////////////////////////////
//  status of shift, control, alt in the Linux console
//  if pressed: shift is 1, control 4, alt 8 - otherwise 0
//  their sum will be returned
//  more precisely: all modifiers defined as power of two
//  in the Linux keymap will be reported
//  the relevant line in the Linux keymap is "keymaps"
//  s. key.c in midnight commander
///////////////////////////////////////////////////////////
{
#if defined __linux__ && !defined X11

    unsigned char modifier=6;

    shiftkey_pressed     = 0;
    ctrlkey_pressed      = 0;
    shiftctrlkey_pressed = 0;
    altkey_pressed       = 0;

    if (ioctl (0, 0x541C, &modifier) < 0) return 0;
    // TIOCLINUX call, function 6:
    // 0 if successful or else -1

    if (modifier > 8) modifier = 0;
    // only shift, control, shiftcontrol, alt
    // other combos not interesting here

    if      (modifier == 1) shiftkey_pressed     = 1;
    else if (modifier == 4) ctrlkey_pressed      = 1;
    else if (modifier == 5) shiftctrlkey_pressed = 1;
    else if (modifier == 8) altkey_pressed       = 1;
    else if (modifier == 2) altkey_pressed       = 1;
    // altgr to be same as alt

    return (int) modifier;

#else

    return 0;

#endif
}

////
int slim_test()
{
    char *kname;
    kname =  CBuf->keymap->name;

    if (NULL != strstr(kname,"slim")) return 1;
    else return 0;
}

////
SLang_Key_Type*
jed_do_keystring (SLKeyMap_List_Type *kml)
{
    register SLang_Key_Type *key, *next, *kmax;
    unsigned int len;
    unsigned char input_ch;
    unsigned char peek_ch=0;
    register unsigned char chup, chlow;
    unsigned char key_ch = 0;
    int q, tmp, kmapswitcher, SLIM=-2;

    SLang_Last_Key_Char = jed_getkey ();
    SLang_Key_TimeOut_Flag = 0;

    if (SLANG_GETKEY_ERROR == (unsigned int) SLang_Last_Key_Char)
        return NULL;

    input_ch = (unsigned char) SLang_Last_Key_Char;

    ///////////////////////////////////////////////////////

    // From here downward modifications
    // to original function from SLang library.
    // Modifications rely on console keymap defaults
    // for all letters a-z (plain and combined).

    //  when executing macro, modifier should come from store
    //  same applies to shift/ctrl/alt key_pressed

    if (Executing_Keyboard_Macro)
    {
        // shift state from Macro_sBuffer_Ptr
        int tmp = * Macro_sBuffer_Ptr;

        modifier             = 0;
        shiftkey_pressed     = 0;
        ctrlkey_pressed      = 0;
        shiftctrlkey_pressed = 0;
        altkey_pressed       = 0;

        if      (tmp == 1) shiftkey_pressed     = 1;
        else if (tmp == 4) ctrlkey_pressed      = 1;
        else if (tmp == 5) shiftctrlkey_pressed = 1;
        else if (tmp == 8) altkey_pressed       = 1;
        else if (tmp == 2) altkey_pressed       = 1;
        // altgr same as alt
    }
    else modifier = get_modifier();

    kmapswitcher = 0;
    if      (shiftkey_pressed)      kmapswitcher = 1;
    else if (ctrlkey_pressed)       kmapswitcher = 4;
    else if (shiftctrlkey_pressed)  kmapswitcher = 5;
    else if (altkey_pressed)        kmapswitcher = 8;

    // for any slim keymap no shift state
    // a keymap is slim if "slim" is in the name
    // but make sure it is a copy of slim_kmap

    if (SINGLE_ESCAPE && input_ch==27)
    {
        // s. unix.c

        tmp = 93;
        ungetkey (&tmp);
        ungetkey (&tmp);
        //  the keystring is now ^[]]
        //  which we call key_escape in the scripts
    }

    if (slim_test())
        kmapswitcher = SLIM;

    char b[20];

    switch (kmapswitcher)
    {
        case 8:
            //  ALT_PRESSED:
            //  exactly two keys are involved
            //  alt+another key
            kml = A_Map;
            altkey_pressed = 0;

            if (input_ch == 27)
            {
                // peek ahead and catch letters,
                // ascii 96-122 in lower case
                // we know that something is there
                // and getkey will not wait for a key press
                peek_ch = jed_getkey ();

                //if (tolower(peek_ch) > 96 && tolower(peek_ch) < 123)
                if (peek_ch != 91)
                {
                    //  not a functional key
                    // next character is a letter
                    // so we got <alt-a>,..., <alt-z>
                    // this is a 128-159 code to ignore
                    // NUL char(0),  + 128 = 128
                    // a is char(1), 1 + 128 = 129
                    // z is char(26), 26 + 128 = 154
                    // 155-159 are practically impossible
                    // // we just discard the previous escape
                    input_ch = tolower (peek_ch); }

                else
                {
                    // next character in keystring is innocent
                    // put it back for processing
                    tmp = (int) peek_ch;
                    if(tmp) ungetkey (&tmp);
                }
            }
            break;

        case 5:
            //  SHIFTCTRL_PRESSED:
            //  exactly three keys are involved
            //  shift+ctrl+another key
            //  namely keys up to space
            //  and keys whose strings begin with escape
            //  and the digit keys 0-9

            if(input_ch < 33 )          kml = S_C_Map;
            else if(input_ch == 127)    kml = S_C_Map;
            else if(input_ch >='0' && input_ch <= '9')
                                        kml = S_C_Map;

            shiftctrlkey_pressed = 0;
            shiftkey_pressed = 1;
            break;

        case 4:
            //  CTRL_PRESSED:
            //  exactly two keys are involved
            //  grey keys whose strings begin with escape

            //  and ctrl+ tab, backspace, return, space
            //  here mostly useless, just in case

            if   (input_ch   == 9)      kml = C_Map;
            else if(input_ch == 13)     kml = C_Map;
            else if(input_ch == 27)     kml = C_Map;
            else if(input_ch == 32)     kml = C_Map;
            else if(input_ch == 127)    kml = C_Map;

            ctrlkey_pressed = 0;
            break;

        case 1:
            //  shift-tab is ESC tab in opensuse, fedora
            //  shift-tab is tab in ubuntu
            if(input_ch==27)
            {
                peek_ch = jed_getkey ();
                if (peek_ch==9) input_ch=peek_ch;
                else
                {
                    tmp = (int) peek_ch;
                    if(tmp) ungetkey (&tmp);
                }
            }

            //  SHIFT_PRESSED:
            //  exactly two keys are involved
            //  shift+another key
            //  namely tab, backspace, return, space
            //  and keys whose strings begin with escape
            if(input_ch == 9)           kml = S_Map;
            else if(input_ch == 13)     kml = S_Map;
            else if(input_ch == 27)     kml = S_Map;
            else if(input_ch == 32)     kml = S_Map;
            else if(input_ch == 127)    kml = S_Map;
            break;

        case -2:
            //  SLIM
            //  any slim keymap as determined in the sl-files
            //  Last_Kbd_Command_String: function

            if (input_ch == 27)
            {
                // Both alt-letter and alt-greykey come here
                // and they must be distinguished.
                // the mess stems from sundry Meta_letter
                // e.g. alt-f generates esc f

                // peek ahead and catch letters,
                // ascii 96-122 in lower case
                // we know that something is there
                // and getkey will not wait for a key press
                peek_ch = jed_getkey ();

                if (
                    (tolower(peek_ch) > 96 && tolower(peek_ch) < 123) ||
                    peek_ch==9)

                {
                    // next character is a letter
                    // so we got <alt-a>,..., <alt-z>
                    // this is an iso 128-159 control to ignore
                    input_ch = tolower (peek_ch);
                }

                else
                {
                    // next character in keystring is not a letter
                    // put it back for processing
                    tmp = (int) peek_ch;
                    if(tmp) ungetkey (&tmp);
                }
            }
            break;

        default:
            kml = Global_Map;
            break;
    }

    CBuf->shift_sel = shiftkey_pressed;

    ///////////////////////////////////////////////////////

    key = (SLang_Key_Type *) &((kml->keymap)[input_ch]);

    if(input_ch > 127 && input_ch < 160)
    {
        key->next == NULL;
        return key;
        //  UTF-8: no key 128-159
    }

    /* if the next one is null, then we know this MAY be it. */
    while (key->next == NULL)
    {
        if (key->type != 0) return key;

        /* Try its opposite case counterpart */
        chlow = tolower (input_ch);
        if (input_ch == chlow) input_ch = toupper (input_ch);
        key = kml->keymap + input_ch;

        if (key->type == 0) return NULL;
    }

    /* It appears to be a prefix character in a key sequence. */

    len = 1;                    /* already read one character */
    key = key->next;            /* Now we are in the key list */
    kmax = NULL;                /* set to end of list */

    while (1)
    {
        SLang_Key_TimeOut_Flag = 1;
        SLang_Last_Key_Char = jed_getkey();
        SLang_Key_TimeOut_Flag = 0;

        len++;

        if ((SLANG_GETKEY_ERROR == (unsigned int) SLang_Last_Key_Char)
        || SLKeyBoard_Quit)
            break;

        input_ch = (unsigned char) SLang_Last_Key_Char;

        chup = toupper (input_ch);
        chlow = tolower (input_ch);

        while (key != kmax)
        {
            if (key->str[0] > len)
            {
                key_ch = key->str[len];
                if (chup == toupper (key_ch)) break;
            }
            key = key->next;
        }

        if (key == kmax) break;

        /* If the input character is lowercase, check to see if there is
         * a lowercase match.  If so, set key to it.  Note: the
         * algorithm assumes the sorting performed by key_string_compare.
         */

        if (input_ch != key_ch)
        {
            next = key->next;
            while (next != kmax)
            {
                if (next->str[0] > len)
                {
                    unsigned char next_ch = next->str[len];
                    if (next_ch == input_ch)
                    {
                        key = next;
                        break;
                    }
                    if (next_ch != chup) break;
                }
                next = next->next;
            }
        }

        /* Ok, we found the first position of a possible match.  If it
         * is exact, we are done.
         */

        if ((unsigned int) key->str[0] == len + 1)
            return key;

        /* There may be ambiguities. Read next key to resolve
         * them. Also adjust kmax to encompass ambiguities.
         */

        next = key->next;
        while (next != kmax)
        {
            if ((unsigned int) next->str[0] > len)
            {
                key_ch = next->str[len];
                if (chup != toupper (key_ch)) break;
            }
            next = next->next;
        }
        kmax = next;
    }

    return NULL;
}

////
int shiftkey_select (void)
{
    //  managed buffer by buffer when switching buffer

    return (CBuf->shift_sel);
}

///////////////////////////////////////////////////////////

static char *find_function_string (FVOID_STAR f)
{
   SLKeymap_Function_Type *fp;

   if (f == (FVOID_STAR) ins_char_cmd) return "self_insert_cmd";

   fp = Jed_Functions;
   while ((fp != NULL) && (fp->name != NULL))
     {
        if ((FVOID_STAR) fp->f == f) return (char *) fp->name;
        fp++;
     }
   return NULL;
}

static SLFUTURE_CONST char *lookup_key_function_string (SLang_Key_Type *key)
{
   if (key == NULL)
     return NULL;

   switch (key->type)
     {
      case SLKEY_F_INTRINSIC:
        return find_function_string (key->f.f);

      case SLKEY_F_INTERPRET:
        return key->f.s;

#ifdef SLKEY_F_SLANG
      case SLKEY_F_SLANG:
        return key->f.slang_fun->name;
#endif
      default:
        return "";
     }
}

static int do_key (void)
{
    SLang_Key_Type *key;
    int repeat;

    if (SLang_Key_TimeOut_Flag == 0)
    {
        Key_Bufferp = Key_Buffer;
        *Key_Bufferp = 0;
    }

    #if JED_HAS_MENUS
    if (Jed_Menus_Active)
    {
        key = jed_menu_do_key ();
        update_jed_keybuffer ();
    }
    else
    #endif
    {
        ////key = SLang_do_key (CBuf->keymap, jed_getkey);
        //// function from SLang library
        //// copied here and replaced with hacked version
        //// to enable modifiers

        if(TXTC)
            key = jed_do_keystring (CBuf->keymap);
        else
            key = SLang_do_key (CBuf->keymap, jed_getkey);
        ////

        update_jed_keybuffer ();
        if ((IN_MINI_WINDOW == 0)
            && (key != NULL)
            && jed_hook_exists ("_jed_before_key_hooks"))
        {
           (void) jed_va_run_hooks ("_jed_before_key_hooks", JED_HOOKS_RUN_ALL,
                                    1, lookup_key_function_string (key));
        }
    }

    if ((key != NULL) && (key->f.f != NULL))
    {
        if (Repeat_Factor == NULL) return key_interpret (key);
        repeat = *Repeat_Factor;
        Suspend_Screen_Update = 1;

        /* some routines may use the repeat factor as a prefix argument */
        while (repeat > 0)
        {
            repeat--;

            if (SLKeyBoard_Quit || SLang_get_error () || (Repeat_Factor == NULL))
                break;

            key_interpret (key);
        }
        Repeat_Factor = NULL;

        return 1;
    }

    else if (!Executing_Keyboard_Macro && !SLKeyBoard_Quit)
    {
        jed_beep ();
        //// key NULL comes here

        flush_input ();
    }

    if (SLKeyBoard_Quit) kbd_quit();
    return 0;
}

int jed_do_key (void)
{
   int ret = do_key ();
   if (SLang_get_error ())
     {
        jed_abort_keyboard_macro ();
        Read_This_Character = NULL;
        Repeat_Factor = NULL;
     }
   return ret;
}

void do_jed(void)
{
   char *mode;
   Buffer *tthis;

   tthis = CBuf;
   mode = tthis->mode_string;

   /* Mark Undo boundary now because this might not be valid later */
   mark_undo_boundary (tthis);

   Repeat_Factor = NULL;
   Replace_Preserve_Case = 0;
   if (jed_do_key()) JWindow->trashed = 1;

    /* internal editing commands may have selected a different buffer
     * so put it back to the one associated with the window.
     */
   if (CBuf != JWindow->buffer)
     {
        char *name;
        if (buffer_exists(JWindow->buffer))
          name = JWindow->buffer->name;
        else name = "*scratch*";
        switch_to_buffer_cmd(name);
     }

#if JED_HAS_MENUS
   if ((tthis != CBuf) || (CBuf->mode_string != mode))
     jed_notify_menu_buffer_changed ();
#endif
}

static int run_switch_active_buffer_hooks (char **bnamep)
{
   char *buffer_name = *bnamep;

   if (buffer_name == CBuf->name)
     return 0;

   if (buffer_name == NULL)
     (void) jed_va_run_hooks ("_jed_switch_active_buffer_hooks", JED_HOOKS_RUN_ALL, 1, "");
   else
     {
        (void) jed_va_run_hooks ("_jed_switch_active_buffer_hooks", JED_HOOKS_RUN_ALL, 1, buffer_name);
        SLang_free_slstring (buffer_name);
     }

   /* This should not fail because CBuf->name is an slstring, and if it
    * does fail, it is not important.
    */
   *bnamep = SLang_create_slstring (CBuf->name);
   return 0;
}

void jed (void)
{
   /* This routine is called from main.  It calls the startup-hooks before
    * entering the main editing loop.
    */
   char *buffer_name;

   if (SLang_get_error ())
     {
        update (NULL, 1, 1, 0);
        SLang_restart (1);
        SLang_set_error (0);
        jed_sleep (2);
     }

#if JED_HAS_MENUS
   jed_notify_menu_buffer_changed ();
#endif

   buffer_name = NULL;
   (void) run_switch_active_buffer_hooks (&buffer_name);
   (void) jed_va_run_hooks ("_jed_startup_hooks", JED_HOOKS_RUN_ALL, 0);
   (void) run_switch_active_buffer_hooks (&buffer_name);

   Jump_Buffer_Ptr = &Jump_Buffer;

   if (setjmp(Jump_Buffer.b) != 0)
     {
        SLang_restart(1);   /* just in case */
     }

   if (CBuf != JWindow->buffer)
     {
        switch_to_buffer(JWindow->buffer);
        window_buffer(CBuf);
     }

#if JED_HAS_MENUS
   jed_notify_menu_buffer_changed ();
#endif

   JWindow->trashed = 1;
   update((Line *) NULL, 0, 0, 1);
   Read_This_Character = NULL;

   while(1)
     {
        int had_err = 0;
        Suspend_Screen_Update = 0;

        do_jed();
        if (SLang_get_error ())
          {
             update (NULL, 1, 0, 0);
             SLang_restart (1);
             SLang_set_error (0);
             had_err = 1;
          }

        if (!SLKeyBoard_Quit && (CBuf->flags & BUFFER_MODIFIED)
            && (!Cursor_Motion)) CBuf->hits += 1;

        SLKeyBoard_Quit = 0;

        if (CBuf->hits > Jed_Max_Hits)
          {
             auto_save_buffer(CBuf);
             check_buffers();   /* check files on disk to see if they are recent */
          }
        (void) run_switch_active_buffer_hooks (&buffer_name);
        if (Jed_Menus_Active == 0)
          (void) jed_va_run_hooks ("_jed_after_key_hooks", JED_HOOKS_RUN_ALL, 0);

        if ((Last_Key_Function != (FVOID_STAR) ins_char_cmd)
            || (JWindow->trashed) || (JWindow != JWindow->next)
            || Suspend_Screen_Update ||
            (Mini_Ghost) || (*Message_Buffer) || (*Error_Buffer))
          {
             update((Line *) NULL, 0, had_err, 1);
          }
     }
}

////int digit_arg(void)

static int which_key (char *f)
{
   int num = 0, i;
   SLang_Key_Type *key, *key_root;
   FVOID_STAR fp;
   unsigned char type;
   unsigned char buf[5];

   if (NULL == (fp = (FVOID_STAR) SLang_find_key_function(f, CBuf->keymap)))
     type = SLKEY_F_INTERPRET;
   else type = SLKEY_F_INTRINSIC;

   i = 256;
   key_root = CBuf->keymap->keymap;
   while (i--)
     {
        key = key_root->next;
        if ((key == NULL) && (type == key_root->type) &&
            (((type == SLKEY_F_INTERPRET) && (!strcmp((char *) f, key_root->f.s)))
             || ((type == SLKEY_F_INTRINSIC) && (fp == key_root->f.f))))
          {
             buf[0] = 2;
             buf[1] = 256 - 1 - i;
             buf[2] = 0;
             SLang_push_string(SLang_make_keystring(buf));
             num++;
          }

        while (key != NULL)
          {
             if ((key->type == type) &&
                 (((type == SLKEY_F_INTERPRET) && (!strcmp((char *) f, key->f.s)))
                  || ((type == SLKEY_F_INTRINSIC) && (fp == key->f.f))))
               {
                  SLang_push_string(SLang_make_keystring(key->str));
                  num++;
               }
             key = key->next;
          }
        key_root++;
     }
   return(num);
}

static void dump_this_binding (SLang_Key_Type *key)
{
   unsigned char ch, *s;
   char *str,  ctrl[2];
   SLFUTURE_CONST char *fun;
   int n, len;

   s = key->str;

   ctrl[0] = '^';
   len = *s++ - 1;;

   while (len-- > 0)
     {
        n = 1;
        ch = *s++;
        if (ch == 127)
          {
             ////str = "DEL"; n = 3;
             str = "Backspace";
             n = strlen(str);
          }
        else if (ch > ' ') str = (char *) &ch;
        else if (ch == 27)
          {
             str = "ESC";
             n = 3;
          }
        else if (ch == ' ')
          {
             str = "SPACE";
             n = 5;
          }
        else if (ch == '\t')
          {
             str = "TAB"; n = 3;
          }
        else
          {
             str = ctrl;
             *(str + 1) = ch + '@';
             n = 2;
          }
        (void) jed_quick_insert((unsigned char *) str, n);
        (void) jed_insert_byte (' ');
     }
   ////(void) jed_insert_wchar_n_times('\t', 3);

    //  this will mostly but not always ensure vertical alignment
    //  just keep sl-function names below 16 characters
    //  combos like ESC UP are not handled properly
    n=17;
    goto_column (&n);

   fun = lookup_key_function_string (key);

   if (fun == NULL) fun = "** Unknown **";
   (void) jed_insert_string (fun);
   (void) jed_insert_newline ();
}

static void dump_bindings(char *map)
{
   int i;
   SLang_Key_Type *next, *key_root;
   SLKeyMap_List_Type *kml;

   CHECK_READ_ONLY_VOID

   if (NULL == (kml = SLang_find_keymap(map)))
     {
        msg_error("Keymap undefined.");
        return;
     }
   key_root = kml->keymap;

   for (i = 0; i < 256; i++)
     {
        next = key_root->next;
        if (next != NULL)
          {
             while (next != NULL)
               {
                  dump_this_binding (next);
                  next = next->next;
               }
          }
        else if (key_root->f.f != NULL) dump_this_binding (key_root);
        key_root++;
     }
}

static void use_keymap(char *name)
{
   SLKeyMap_List_Type *map;

   if ((name == NULL) || (*name == 0)
       || (NULL == (map = SLang_find_keymap(name))))
     {
        msg_error("Unknown keymap.");
     }
   else CBuf->keymap = map;
}

static SLFUTURE_CONST char *what_keymap (void)
{
   return CBuf->keymap->name;
}

////void jed_set_abort_char (int *c)

static SLKeymap_Function_Type *Flist_Context;
static int Flist_Context_Len;

static char *Slang_Functions[JED_MAX_ADD_COMPLETIONS];
static char **Slang_Flist_Context;

int next_function_list(char *buf)
{
   char *name;
   char **max;

   /* Convert '-' to '_' */

   name = buf;
   while (*name != 0)
     {
        if (*name == '-') *name = '_';
        name++;
     }

   while (1)
     {
        SLKeymap_Function_Type *tthis = Flist_Context;
        SLFUTURE_CONST char *tname = tthis->name;
        if (tname == NULL) break;
        Flist_Context++;
        if (!Flist_Context_Len || !strncmp(buf, tname, Flist_Context_Len))
          {
             strcpy(buf, tname);
             return(1);
          }
     }

   max = Slang_Functions + JED_MAX_ADD_COMPLETIONS;
   while (Slang_Flist_Context < max)
     {
        name = *Slang_Flist_Context;
        if (name == NULL) return(0);
#if 0
        name++;  /* skip past hash mark */
#endif
        Slang_Flist_Context++;
        if (!Flist_Context_Len || !strncmp(buf, name, Flist_Context_Len))
          {
             strcpy(buf, name);
             return(1);
          }
     }
   return(0);
}

int open_function_list(char *buf)
{
   Flist_Context = Jed_Functions;
   Slang_Flist_Context = Slang_Functions;
   Flist_Context_Len = strlen(buf);
   return next_function_list(buf);
}

int add_to_completion(char *name)
{
   static char **last = Slang_Functions;

   if (last == Slang_Functions + JED_MAX_ADD_COMPLETIONS)
     return 0;

   if (0 == SLang_is_defined (name))
     {
        jed_verror ("add_completion: %s undefined", name);
        return -1;
     }

   name = SLang_create_slstring (name);
   if (name == NULL)
     return -1;

   *last++ = name;
   return 0;
}

int is_internal(char *f)
{
   if (NULL == SLang_find_key_function(f, CBuf->keymap)) return(0);
   return 1;
}

static void copy_keymap_cmd (char *dest, char *src)
{
   SLKeyMap_List_Type *ksrc;

   ksrc = SLang_find_keymap (src);
   if (ksrc == NULL)
     {
        jed_verror ("Keymap \"%s\" does not exist",  src);
        return;
     }
   if (NULL != SLang_find_keymap (dest))
     {
        jed_verror ("Keymap \"%s\" already exists", dest);
        return;
     }

   if (NULL == SLang_create_keymap (dest, ksrc))
     jed_verror ("Unable to create keymap \"%s\"", dest);
}

void jed_call_cmd (char *str)
{
   SLKeyMap_List_Type *km;
   int (*fp)(void);

   if (*str == ' ')
     {
        jed_insert_string (str + 1);
        return;
     }

   if (*str == '@')
     {
        (void) do_macro_string (str + 1, 1);
        return;
     }

   km = CBuf->keymap;
   if (NULL == (fp = (int (*)(void)) (SLang_find_key_function(str, km))))
     jed_verror ("Internal function %s does not exist", str);
   else (void) (*fp)();
}

typedef struct
{
   unsigned char *key;
   unsigned int len;
   unsigned int num_read;
}
Getkey_Callback_Data_Type;

static Getkey_Callback_Data_Type Getkey_Callback_Data;

static int get_key_function_callback (void)
{
   unsigned int i = Getkey_Callback_Data.num_read;

   if (i == Getkey_Callback_Data.len)
     return SLANG_GETKEY_ERROR;

   Getkey_Callback_Data.num_read += 1;
   return (int) Getkey_Callback_Data.key[i];
}

// not used at all
static int push_key_binding (unsigned char *keystr)
{
   SLang_Key_Type *key;
   char *fun;
   int type;
#ifdef SLKEY_F_SLANG
   SLang_Name_Type *ref = NULL;
#endif

   SLang_Key_TimeOut_Flag = 0;

   if (keystr == NULL)
     {
        Key_Bufferp = Key_Buffer;
        key = SLang_do_key (CBuf->keymap, jed_getkey);
        update_jed_keybuffer ();
        if (key == NULL)
          flush_input ();

        if (SLKeyBoard_Quit || (SLang_get_error () == USER_BREAK))
          {
             SLang_set_error (0);
             SLKeyBoard_Quit = 0;
          }
     }
   else
     {
        Getkey_Callback_Data.len = (unsigned int) keystr[0];
        Getkey_Callback_Data.key = keystr + 1;
        Getkey_Callback_Data.num_read = 0;
        key = SLang_do_key (CBuf->keymap, get_key_function_callback);
     }

   fun = NULL;
   type = -1;

   if (key != NULL)
     switch (key->type)
       {
        case SLKEY_F_INTRINSIC:
          type = 1;
          fun = find_function_string (key->f.f);
          break;

        case SLKEY_F_INTERPRET:
          fun = key->f.s;
          type = 0;
          if (fun[0] == '@') type = 2;
          else if (fun[0] == ' ') type = 3;
          break;

#ifdef SLKEY_F_SLANG
        case SLKEY_F_SLANG:
          type = 4;
          ref = key->f.slang_fun;
          break;
#endif
     }

   if (-1 == SLang_push_integer (type))
     return -1;

#ifdef SLKEY_F_SLANG
   if (type == 4)
     type = SLang_push_function (ref);
   else
#endif
     type = SLang_push_string (fun);

   return type;
}

////  used only once below
int jed_get_key_function (unsigned char *keystr, char **fun)
{
    SLang_Key_Type *key;

    SLang_Key_TimeOut_Flag = 0;

    *fun = NULL;

    if (keystr == NULL)
    // case where get_key_binding has no parameter
    {
        Key_Bufferp = Key_Buffer;

        //key = SLang_do_key (CBuf->keymap, jed_getkey);
        //key = jed_do_keystring (CBuf->keymap, jed_getkey);
        key = jed_do_keystring (CBuf->keymap);

        update_jed_keybuffer ();
        if (key == NULL) flush_input ();

        if (SLKeyBoard_Quit && (SLang_get_error () == USER_BREAK))
        {
            SLang_set_error (0);
            SLKeyBoard_Quit = 0;
        }
    }

    // needs amendment for shiftstate
    // never ever used, comment out?
    else
    {
        Getkey_Callback_Data.len = (unsigned int) keystr[0];
        Getkey_Callback_Data.key = keystr + 1;
        Getkey_Callback_Data.num_read = 0;
        key = SLang_do_key (CBuf->keymap, get_key_function_callback);
        //
    }

    if (key == NULL) return -1;

    if (key->type == SLKEY_F_INTRINSIC)
    {
        *fun = find_function_string (key->f.f);
        return 1;
    }

    else
    {
        *fun = key->f.s;
        if (key->f.s[0] == '@') return 2;
        if (key->f.s[0] == ' ') return 3;
        return 0;
    }
}

////static void get_key_binding (void)

////    custom
static void get_key_binding (void)
{
   unsigned char *s = NULL;
   char *sstr;
   int type;
   //

   if (SLang_Num_Function_Args == 1)
     {
        char *str;

        if (-1 == SLang_pop_slstring (&str))
          return;

        s = (unsigned char *) SLang_process_keystring (str);
        SLang_free_slstring (str);

        if (s == NULL)
          return;
     }

    // key press determined string
    // s is NULL if no parameter
    type = jed_get_key_function (s, &sstr);

    // modifier is available here
    if (type>=0) type = 10*modifier + type;

    (void) SLang_push_integer (type);

    //(void) push_key_binding (s);
    // should be use instead of SLang_push_string
    // type and modifier should move there
    // it runs fine in conjunction with display_cmd, suemenu.sl

    (void) SLang_push_string (sstr);
}

/* This is weird, Ultrix cc will not compile if set_key comes before unset_key */
static void unset_key (char *key)
{
   if (*key)
     SLang_undefine_key(key, Global_Map);
}

static void unset_key_in_keymap(char *key, char *map)
{
   SLKeyMap_List_Type *kmap;

   if (NULL == (kmap = SLang_find_keymap(map)))
     {
        jed_verror ("Unknown keymap: %s", map);
        return;
     }

   if (*key) SLang_undefine_key(key, kmap);
}

static void set_key_in_keymap_1 (SLKeyMap_List_Type *kmap)
{
   char *key, *func;

   if (-1 == SLang_pop_slstring (&key))
     return;
#ifdef SLKEY_F_SLANG
   if (SLang_peek_at_stack () == SLANG_REF_TYPE)
     {
        SLang_Name_Type *nt = SLang_pop_function ();
        if (nt == NULL)
          {
             SLang_free_slstring (key);
             return;
          }
        if (-1 == SLkm_define_slkey (key, nt, kmap))
          {
             SLang_free_function (nt);
             SLang_free_slstring (key);
             return;
          }
        SLang_free_slstring (key);
        return;
     }
#endif
   if (-1 == SLang_pop_slstring (&func))
     {
        SLang_free_slstring (key);
        return;
     }

   if (*key)
     (void) SLang_define_key(key, func, kmap);

   SLang_free_slstring (func);
   SLang_free_slstring (key);
}

static void set_key_in_keymap (void)
{
   char *keymap;
   SLKeyMap_List_Type *kmap;

   if (-1 == SLang_pop_slstring (&keymap))
     return;

   if (NULL == (kmap = SLang_find_keymap(keymap)))
     jed_verror ("Unknown keymap: %s", keymap);
   else
     set_key_in_keymap_1 (kmap);

   SLang_free_slstring (keymap);
}

static void set_key (void)
{
   set_key_in_keymap_1 (Global_Map);
}

static int keymap_p(char *name)
{
   return ! (NULL == SLang_find_keymap(name));
}

static void make_keymap_cmd (char *km)
{
   copy_keymap_cmd (km, "global");
}

static void get_keymaps (void)
{
   int j, n;
   char **keymaps;
   SLang_Array_Type *at;
#if SLANG_VERSION < 20000
   unsigned int i;
#else
   SLkeymap_Type *km;
#endif

   n = 0;
#if SLANG_VERSION < 20000
   for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
     {
        if (SLKeyMap_List[i].keymap == NULL)
          continue;

        n++;
     }
#else
   km = SLKeyMap_List_Root;
   while (km != NULL)
     {
        n++;
        km = km->next;
     }
#endif

   if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &n, 1)))
     return;

   keymaps = (char **)at->data;
   j = 0;
#if SLANG_VERSION < 20000
   for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
     {
        if (SLKeyMap_List[i].keymap == NULL)
          continue;

        if (j == n)
          break;                       /* keymap list changing?? */

        if (NULL == (keymaps[j] = SLang_create_slstring (SLKeyMap_List[i].name)))
          {
             SLang_free_array (at);
             return;
          }
        j++;
     }
#else
   km = SLKeyMap_List_Root;
   while (km != NULL)
     {
        if (j == n)
          break;                       /* keymap list changing?? */

        if (NULL == (keymaps[j] = SLang_create_slstring (km->name)))
          {
             SLang_free_array (at);
             return;
          }
        km = km->next;
        j++;
     }
#endif
   SLang_push_array (at, 1);
}

static void getkey_wchar_intrin (void)
{
   int ch;
   SLwchar_Type wch;

   if (-1 == jed_getkey_wchar (&wch))
     {
        ch = (int) wch;
        (void) SLang_push_integer (-ch);
        return;
     }
   (void) SLclass_push_long_obj (SLANG_LONG_TYPE, (long)wch);
}

static void ungetkey_wchar_intrin (long *chp)
{
   if (*chp < 0)
     {
        int ch = -(int)*chp;
        ungetkey (&ch);
        return;
     }
   jed_ungetkey_wchar ((SLwchar_Type)*chp);
}

static SLang_Intrin_Fun_Type Keymap_Intrinsics [] =
{
   MAKE_INTRINSIC_0("setkey", set_key, VOID_TYPE),
   MAKE_INTRINSIC_0("definekey", set_key_in_keymap, VOID_TYPE),
   MAKE_INTRINSIC_S("unsetkey", unset_key, VOID_TYPE),
   MAKE_INTRINSIC("what_keymap", what_keymap, STRING_TYPE, 0),
   MAKE_INTRINSIC_0("get_keymaps", get_keymaps, VOID_TYPE),
   MAKE_INTRINSIC_S("keymap_p",  keymap_p, INT_TYPE),
   MAKE_INTRINSIC("get_key_binding", get_key_binding , VOID_TYPE, 0),
   MAKE_INTRINSIC_S("dump_bindings", dump_bindings, VOID_TYPE),
   MAKE_INTRINSIC_SS("undefinekey", unset_key_in_keymap, VOID_TYPE),
   MAKE_INTRINSIC_S("which_key", which_key, INT_TYPE),
   MAKE_INTRINSIC_S("make_keymap", make_keymap_cmd, VOID_TYPE),
   MAKE_INTRINSIC_SS("copy_keymap", copy_keymap_cmd, VOID_TYPE),
   MAKE_INTRINSIC_S("use_keymap",  use_keymap, VOID_TYPE),
   MAKE_INTRINSIC("_getkey", jed_getkey, INT_TYPE, 0),
   MAKE_INTRINSIC_I("_ungetkey", ungetkey, VOID_TYPE),
   MAKE_INTRINSIC_0("getkey", getkey_wchar_intrin, VOID_TYPE),
   MAKE_INTRINSIC_1("ungetkey", ungetkey_wchar_intrin, VOID_TYPE, SLANG_LONG_TYPE),
#ifdef REAL_UNIX_SYSTEM
   MAKE_INTRINSIC_1("set_default_key_wait_time", jed_set_default_key_wait_time, SLANG_INT_TYPE, SLANG_INT_TYPE),
#endif
   SLANG_END_INTRIN_FUN_TABLE
};

static char *Key_Buffer_Ptr = Jed_Key_Buffer;
static char *Last_Kbd_Command_Str_Ptr = Last_Kbd_Command_String;
static char *Current_Kbd_Command_Str_Ptr = Current_Kbd_Command_String;

static SLang_Intrin_Var_Type Keymap_Variables [] =
{
   MAKE_VARIABLE("LASTKEY", &Key_Buffer_Ptr, STRING_TYPE, 1),
   MAKE_VARIABLE("LAST_KBD_COMMAND", &Last_Kbd_Command_Str_Ptr, STRING_TYPE, 1),
   MAKE_VARIABLE("CURRENT_KBD_COMMAND", &Current_Kbd_Command_Str_Ptr, STRING_TYPE, 1),
   MAKE_VARIABLE(NULL, NULL, 0, 0)
};

int jed_init_keymap_intrinsics (void)
{
   if ((-1 == SLadd_intrin_fun_table (Keymap_Intrinsics, NULL))
       || (-1 == SLadd_intrin_var_table (Keymap_Variables, NULL)))
     return -1;

   return 0;
}
